/* schemes-entry-row.c
 *
 * Copyright 2021 Christian Hergert <chergert@redhat.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#include "config.h"

#include <glib/gi18n.h>

#include "schemes-entry-row.h"

struct _SchemesEntryRow
{
  AdwPreferencesRow  parent_instance;
  GtkText           *text;
  GtkLabel          *label;
};

static void editable_iface_init (GtkEditableInterface *iface);

G_DEFINE_TYPE_WITH_CODE (SchemesEntryRow, schemes_entry_row, ADW_TYPE_PREFERENCES_ROW,
                         G_IMPLEMENT_INTERFACE (GTK_TYPE_EDITABLE, editable_iface_init))

static guint activate_signal;

GtkWidget *
schemes_entry_row_new (void)
{
  return g_object_new (SCHEMES_TYPE_ENTRY_ROW, NULL);
}

static gboolean
schemes_entry_row_grab_focus (GtkWidget *widget)
{
  return gtk_widget_grab_focus (GTK_WIDGET (SCHEMES_ENTRY_ROW (widget)->text));
}

static void
on_text_activate_cb (SchemesEntryRow *self,
                     GtkText           *text)
{
  g_assert (SCHEMES_IS_ENTRY_ROW (self));
  g_assert (GTK_IS_TEXT (text));

  g_signal_emit (self, activate_signal, 0);
}

static void
on_text_notify_cb (SchemesEntryRow *self,
                   GParamSpec        *pspec,
                   GtkText           *text)
{
  GObjectClass *klass;

  g_assert (SCHEMES_IS_ENTRY_ROW (self));
  g_assert (GTK_IS_TEXT (text));

  klass = G_OBJECT_GET_CLASS (self);
  pspec = g_object_class_find_property (klass, pspec->name);

  if (pspec != NULL)
    g_object_notify_by_pspec (G_OBJECT (self), pspec);
}

static void
schemes_entry_row_dispose (GObject *object)
{
  SchemesEntryRow *self = (SchemesEntryRow *)object;

  self->text = NULL;
  self->label = NULL;

  G_OBJECT_CLASS (schemes_entry_row_parent_class)->dispose (object);
}

static void
schemes_entry_row_get_property (GObject    *object,
                                guint       prop_id,
                                GValue     *value,
                                GParamSpec *pspec)
{
  if (gtk_editable_delegate_get_property (object, prop_id, value, pspec))
    return;

  G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
}

static void
schemes_entry_row_set_property (GObject      *object,
                                guint         prop_id,
                                const GValue *value,
                                GParamSpec   *pspec)
{
  if (gtk_editable_delegate_set_property (object, prop_id, value, pspec))
    return;

  G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
}

static void
schemes_entry_row_class_init (SchemesEntryRowClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

  object_class->dispose = schemes_entry_row_dispose;
  object_class->get_property = schemes_entry_row_get_property;
  object_class->set_property = schemes_entry_row_set_property;

  widget_class->grab_focus = schemes_entry_row_grab_focus;

  gtk_editable_install_properties (object_class, 1);

  activate_signal = gtk_widget_class_get_activate_signal (widget_class);

  gtk_widget_class_set_layout_manager_type (widget_class, GTK_TYPE_BOX_LAYOUT);
  gtk_widget_class_set_accessible_role (widget_class, GTK_ACCESSIBLE_ROLE_TEXT_BOX);
  gtk_widget_class_set_template_from_resource (widget_class, "/ui/schemes-entry-row.ui");
  gtk_widget_class_bind_template_child (widget_class, SchemesEntryRow, label);
  gtk_widget_class_bind_template_child (widget_class, SchemesEntryRow, text);
  gtk_widget_class_bind_template_callback (widget_class, on_text_activate_cb);
  gtk_widget_class_bind_template_callback (widget_class, on_text_notify_cb);
}

static void
schemes_entry_row_init (SchemesEntryRow *self)
{
  gtk_widget_init_template (GTK_WIDGET (self));

  g_object_bind_property (self, "title", self->label, "label", G_BINDING_SYNC_CREATE);
  g_object_bind_property (self, "use-underline", self->label, "use-underline", G_BINDING_SYNC_CREATE);
}

static GtkEditable *
schemes_entry_row_get_delegate (GtkEditable *editable)
{
  return GTK_EDITABLE (SCHEMES_ENTRY_ROW (editable)->text);
}

static void
editable_iface_init (GtkEditableInterface *iface)
{
  iface->get_delegate = schemes_entry_row_get_delegate;
}
